from inewave._utils.arquivo import Arquivo
from inewave._utils.dadosarquivo import DadosArquivo
from inewave.nwlistop.modelos.earmfpsin import LeituraEarmfpSIN

import pandas as pd  # type: ignore


class EarmfpSIN(Arquivo):
    """
    Armazena os dados das saídas referentes às energias
    armazenadas finais, por submercado e em % da energia armazenável máxima.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `earmfpm00x.out`, onde x varia conforme o
    submercado em questão.
    """
    def __init__(self,
                 dados: DadosArquivo):
        super().__init__(dados)

    # Override
    @classmethod
    def le_arquivo(cls,
                   diretorio: str,
                   nome_arquivo="earmfpsin.out") -> 'EarmfpSIN':
        """
        """
        leitor = LeituraEarmfpSIN(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    @property
    def energias(self) -> pd.DataFrame:
        """
        Tabela com as energias armazenadas percentuais por série e
        por mês/ano de estudo.

         **Retorna**

        `pd.DataFrame`
        """
        return self._blocos[0].dados



# Rotinas de testes associadas ao arquivo adterm.dat do NEWAVE
from inewave.newave.sistema import Sistema


sis = Sistema.le_arquivo("tests/_arquivos")


def test_leitura():
    assert len(sis.custo_deficit) > 0


def test_escrita_e_leitura():
    sis.escreve_arquivo("tests/_saidas")
    sis2 = Sistema.le_arquivo("tests/_saidas")
    assert sis == sis2


def test_eq_sistema():
    sis2 = Sistema.le_arquivo("tests/_arquivos")
    assert sis == sis2


def test_neq_sistema():
    sis2 = Sistema.le_arquivo("tests/_arquivos")
    sis2.custo_deficit.iloc[0, 0] = -1
    assert sis2 != sis
    assert sis2 is not None



# Rotinas de testes associadas ao arquivo adterm.dat do NEWAVE
from inewave.newave.curva import Curva


cur = Curva.le_arquivo("tests/_arquivos")


def test_leitura():
    assert len(cur.curva_seguranca) > 0


def test_escrita_e_leitura():
    cur.escreve_arquivo("tests/_saidas")
    cur2 = Curva.le_arquivo("tests/_saidas")
    assert cur == cur2


def test_eq_curva():
    cur2 = Curva.le_arquivo("tests/_arquivos")
    assert cur == cur2


def test_neq_curva():
    cur2 = Curva.le_arquivo("tests/_arquivos")
    cur2.curva_seguranca.iloc[0, 0] = -1
    assert cur2 != cur
    assert cur2 is not None



# Imports do próprio módulo
from inewave._utils.bloco import Bloco
from inewave._utils.leitura import Leitura
from inewave._utils.registros import RegistroAn
# Imports de módulos externos
from typing import IO, List


class BlocoNomesArquivos(Bloco):
    """
    Bloco de informações do arquivo de
    entrada do NEWAVE `arquivos.dat`.
    """

    legendas = [
                "DADOS GERAIS                :",
                "DADOS DOS SUBSIST/SUBMERCADO:",
                "CONFIGURACAO HIDRAULICA     :",
                "ALTERACAO DADOS USINAS HIDRO:",
                "CONFIGURACAO TERMICA        :",
                "DADOS DAS USINAS TERMICAS   :",
                "DADOS DAS CLASSES TERMICAS  :",
                "DADOS DE EXPANSAO HIDRAULICA:",
                "ARQUIVO DE EXPANSAO TERMICA :",
                "ARQUIVO DE PATAMARES MERCADO:",
                "ARQUIVO DE CORTES DE BENDERS:",
                "ARQUIVO DE CABECALHO CORTES :",
                "RELATORIO DE CONVERGENCIA   :",
                "RELATORIO DE E. SINTETICAS  :",
                "RELATORIO DETALHADO FORWARD :",
                "ARQUIVO DE CABECALHO FORWARD:",
                "ARQUIVO DE S.HISTORICAS S.F.:",
                "ARQUIVO DE MANUT.PROG. UTE'S:",
                "ARQUIVO P/DESPACHO HIDROTERM:",
                "ARQUIVO C/TEND. HIDROLOGICA :",
                "ARQUIVO C/DADOS DE ITAIPU   :",
                "ARQUIVO C/DEMAND S. BIDDING :",
                "ARQUIVO C/CARGAS ADICIONAIS :",
                "ARQUIVO C/FATORES DE PERDAS :",
                "ARQUIVO C/PATAMARES GTMIN   :",
                "ARQUIVO ENSO 1              :",
                "ARQUIVO ENSO 2              :",
                "ARQUIVO DSVAGUA             :",
                "ARQUIVO P/PENALID. POR DESV.:",
                "ARQUIVO C.GUIA / PENAL.VMINT:",
                "ARQUIVO AGRUPAMENTO LIVRE   :",
                "ARQUIVO ANTEC.DESP.GNL      :",
                "ARQUIVO GER. HIDR. MIN.     :",
                "ARQUIVO AVERSAO RISCO - SAR :",
                "ARQUIVO AVERSAO RISCO - CVAR:",
                "DADOS DOS RESER.EQ.ENERGIA  :",
                "DADOS DOS REST. ELETRICAS   :",
                "ARQUIVO DE TECNOLOGIAS      :",
                "DADOS DE ABERTURAS          :",
                "ARQUIVO DE EMISSOES GEE     :",
                "ARQUIVO DE RESTRICAO DE GAS :"
               ]

    def __init__(self,
                 str_inicio: str,
                 str_final: str,
                 obrigatorio: bool):

        super().__init__(str_inicio,
                         str_final,
                         obrigatorio)

        self._dados = [""] * len(BlocoNomesArquivos.legendas)

    # Override
    def le(self, arq: IO):
        reg = RegistroAn(12)
        self._dados[0] = reg.le_registro(self._linha_inicio, 30)
        for i in range(len(self._dados)):
            linha = arq.readline()
            # Confere se já terminou (possíveis \n ao final)
            if len(linha) < 3:
                break
            reg = RegistroAn(12)
            self._dados[i + 1] = reg.le_registro(linha, 30)

    # Override
    def escreve(self, arq: IO):
        for leg, nome in zip(BlocoNomesArquivos.legendas,
                             self._dados):
            arq.write(f"{leg} {nome}\n")


class LeituraArquivos(Leitura):
    """
    Realiza a leitura do arquivo `arquivos.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `arquivos.dat`, construindo
    um objeto `Arquivos` cujas informações são as mesmas do arquivos.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self,
                 diretorio: str):
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Cria a lista de blocos a serem lidos no arquivo arquivos.dat.
        """
        nomes_arquivos = BlocoNomesArquivos("", "", True)
        return [nomes_arquivos]



from inewave.config import REES, MAX_ANOS_ESTUDO, MESES_DF
from inewave._utils.registros import RegistroIn, RegistroFn
from inewave._utils.bloco import Bloco
from inewave._utils.leitura import Leitura

from typing import List, IO
import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class BlocoConfiguracoesPenalizacaoCurva(Bloco):
    """
    Bloco com as informações de configurações de
    penalização da curva de volume mínimo.
    """
    str_inicio = ""
    str_fim = ""

    def __init__(self):

        super().__init__(BlocoConfiguracoesPenalizacaoCurva.str_inicio,
                         "",
                         True)

        self._dados: np.ndarray = np.zeros((3,),
                                           dtype=np.int64)

    def __eq__(self, o: object):
        if not isinstance(o, BlocoConfiguracoesPenalizacaoCurva):
            return False
        bloco: BlocoConfiguracoesPenalizacaoCurva = o
        return np.array_equal(self._dados, bloco._dados)

    # Override
    def le(self, arq: IO):
        reg = RegistroIn(3)
        linha = arq.readline()
        self._dados = np.array(reg.le_linha_tabela(linha, 1, 1, 3),
                               dtype=np.int64)
        return

    # Override
    def escreve(self, arq: IO):
        cab = (" XXX XXX XXX (TIPO DE PENALIZACAO DAS VIOLACOES: 0-FIXA" +
               " 1-MAXIMA; MES DE PENALIZACAO;VMINOP SAZONAL NO PRE/POS:" +
               " 0-NAO CONSIDERA 1-CONSIDERA)")
        arq.write(cab + "\n")
        linha = ""
        for d in self._dados:
            linha += f" {str(d).rjust(3)}"
        arq.write(linha + "\n")


class BlocoPenalidadesViolacaoREECurva(Bloco):
    """
    Bloco com informações das penalidades por violação para cada
    ree.
    """
    str_inicio = "SISTEMA   CUSTO"
    str_fim = "999"

    def __init__(self):

        super().__init__(BlocoPenalidadesViolacaoREECurva.str_inicio,
                         "",
                         True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoPenalidadesViolacaoREECurva):
            return False
        bloco: BlocoPenalidadesViolacaoREECurva = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):

        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            df.columns = ["Subsistema", "Custo"]
            df = df.astype({"Subsistema": 'int32'})
            return df

        # Variáveis auxiliares
        reg_subsis = RegistroIn(3)
        reg_custo = RegistroFn(7)
        # Pula uma linha, com cabeçalhos
        arq.readline()
        i = 0
        tabela = np.zeros((len(REES),
                          2))
        while True:
            # Verifica se o arquivo acabou
            linha: str = arq.readline()
            if (BlocoPenalidadesViolacaoREECurva.str_fim
               == linha.strip()):
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            tabela[i, 0] = reg_subsis.le_registro(linha, 1)
            tabela[i, 1] = reg_custo.le_registro(linha, 11)
            i += 1

    # Override
    def escreve(self, arq: IO):

        def escreve_custos():
            lin_tab = self._dados.shape[0]
            for i in range(lin_tab):
                linha = f" {str(int(self._dados.iloc[i, 0])).zfill(3)}"
                linha += f"       {self._dados.iloc[i, 1]:4.2f}"
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        arq.write(f" {BlocoPenalidadesViolacaoREECurva.str_inicio}\n")
        arq.write(" XXX       XXXX.XX\n")
        escreve_custos()
        # Escreve a linha de terminação
        arq.write(f" {BlocoPenalidadesViolacaoREECurva.str_fim}\n")


class BlocoCurvaSegurancaSubsistema(Bloco):
    """
    Bloco com informações da curva de segurança de operação por mês/ano
    e por subsistema.
    """
    str_inicio = "CURVA DE SEGURANCA"
    str_fim = "9999"

    def __init__(self):

        super().__init__(BlocoCurvaSegurancaSubsistema.str_inicio,
                         "",
                         True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoCurvaSegurancaSubsistema):
            return False
        bloco: BlocoCurvaSegurancaSubsistema = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):

        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            df.columns = MESES_DF
            df["Ano"] = anos
            df["REE"] = subsistema
            df = df[["Ano", "REE"] + MESES_DF]
            return df

        # Variáveis auxiliares
        reg_subsis = RegistroIn(3)
        reg_ano = RegistroIn(4)
        reg_curva = RegistroFn(5)
        # Pula duas linhas, com cabeçalhos
        arq.readline()
        arq.readline()
        i = 0
        anos = []
        subsistema = []
        subsistema_atual = 0

        tabela = np.zeros((MAX_ANOS_ESTUDO * len(REES),
                          len(MESES_DF)))
        while True:
            # Verifica se o arquivo acabou
            linha: str = arq.readline()
            if BlocoCurvaSegurancaSubsistema.str_fim == linha.strip():
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            # Senão, lê mais uma linha
            if len(linha.strip()) < 10:
                subsistema_atual = reg_subsis.le_registro(linha, 1)
            else:
                # Ano
                anos.append(reg_ano.le_registro(linha, 0))
                # Subsistema
                subsistema.append(subsistema_atual)
                # Limites
                tabela[i, :] = reg_curva.le_linha_tabela(linha,
                                                         6,
                                                         1,
                                                         len(MESES_DF))
                i += 1

    # Override
    def escreve(self, arq: IO):

        def escreve_curva():
            lin_tab = self._dados.shape[0]
            subsistema_anterior = 0
            for i in range(lin_tab):
                linha = ""
                # Subsistema
                subsistema = self._dados.iloc[i, 1]
                if subsistema != subsistema_anterior:
                    subsistema_anterior = subsistema
                    linha = " " + str(subsistema).rjust(3)
                    arq.write(linha + "\n")
                    subsistema_anterior = subsistema
                # Curva de cada mês
                linha = f"{str(self._dados.iloc[i, 0]).ljust(4)} "
                for j in range(len(MESES_DF)):
                    v = self._dados.iloc[i, j + 2]
                    linha += " " + "{:3.1f}".format(v).rjust(5)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        arq.write(f" {BlocoCurvaSegurancaSubsistema.str_inicio}\n")
        arq.write(" XXX\n")
        cab = ("      JAN.X FEV.X MAR.X ABR.X MAI.X JUN.X JUL.X" +
               " AGO.X SET.X OUT.X NOV.X DEZ.X\n")
        arq.write(cab)
        escreve_curva()
        # Escreve a linha de terminação
        arq.write(f"{BlocoCurvaSegurancaSubsistema.str_fim}\n")


class LeituraCurva(Leitura):
    """
    Realiza a leitura do arquivo `curva.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `curva.dat`, construindo
    um objeto `Curva` cujas informações são as mesmas do curva.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self,
                 diretorio: str):
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Cria a lista de blocos a serem lidos no arquivo curva.dat.
        """
        return [BlocoConfiguracoesPenalizacaoCurva(),
                BlocoPenalidadesViolacaoREECurva(),
                BlocoCurvaSegurancaSubsistema()]



from inewave._utils.bloco import Bloco
from inewave._utils.leitura import Leitura
from inewave._utils.registros import RegistroAn, RegistroIn, RegistroFn
from inewave.config import MAX_LAG_ADTERM, MAX_UTES, NUM_PATAMARES

from typing import List, IO
import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class BlocoUTEsAdTerm(Bloco):
    """
    Bloco com os despachos antecipados das UTEs por patamar.
    """
    str_inicio = ""
    str_fim = "9999"

    def __init__(self):

        super().__init__(BlocoUTEsAdTerm.str_inicio,
                         "",
                         True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoUTEsAdTerm):
            return False
        bloco: BlocoUTEsAdTerm = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):

        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            pats = [n for n in range(1, NUM_PATAMARES + 1)]
            cols_pats = [f"Patamar {p}" for p in pats]
            df.columns = cols_pats
            df["Índice UTE"] = iutes
            df["Nome"] = nomes_utes
            df["Lag"] = lags_utes
            df = df[["Índice UTE", "Nome", "Lag"] + cols_pats]
            return df

        # Variáveis auxiliares
        reg_iute = RegistroIn(4)
        reg_nome = RegistroAn(12)
        reg_despacho = RegistroFn(10)
        # Pula as duas primeiras linhas, com cabeçalhos
        arq.readline()
        i = 0
        iutes = []
        nomes_utes = []
        lags_utes = []
        iute_atual = 0
        nome_ute_atual = ""
        lag_leitura = 1
        tabela = np.zeros((MAX_UTES * MAX_LAG_ADTERM,
                           NUM_PATAMARES))
        while True:
            # Verifica se o arquivo acabou
            linha: str = arq.readline()
            if BlocoUTEsAdTerm.str_fim in linha:
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            # Senão, lê mais uma linha
            # Ano
            if linha[1:5].strip().isnumeric():
                iute_atual = reg_iute.le_registro(linha, 1)
                nome_ute_atual = reg_nome.le_registro(linha, 7)
                lag_leitura = 1
            else:
                iutes.append(iute_atual)
                nomes_utes.append(nome_ute_atual)
                lags_utes.append(lag_leitura)
                # Patamares
                tabela[i, :] = reg_despacho.le_linha_tabela(linha,
                                                            24,
                                                            2,
                                                            NUM_PATAMARES)
                lag_leitura += 1
                i += 1

    # Override
    def escreve(self, arq: IO):

        def escreve_termicas():
            tabela = self._dados
            lin_tab = tabela.shape[0]
            for i in range(lin_tab):
                iute = tabela.iloc[i, 0]
                nome = tabela.iloc[i, 1]
                lag = tabela.iloc[i, 2]
                if lag == 1:
                    # Descobre o número de lags da UTE
                    lag_ute = tabela.loc[tabela["Índice UTE"] == iute,
                                         "Índice UTE"].shape[0]
                    linha_ute = (f" {str(iute).rjust(4)}  " +
                                 f"{str(nome).ljust(12)}  {lag_ute}")
                    arq.write(linha_ute + "\n")

                # Despachos
                linha = "                      "
                despachos = len([c for c in list(self._dados.columns)
                                 if "Patamar" in c])
                for j in range(despachos):
                    v = tabela.iloc[i, j + 3]
                    linha += "  " + "{:7.2f}".format(v).rjust(10)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        titulos = " IUTE  NOME TERMICA LAG" + "\n"
        cab = (" XXXX  XXXXXXXXXXXX  X  XXXXXXX.XX" +
               "  XXXXXXX.XX  XXXXXXX.XX" + "\n")
        arq.write(titulos)
        arq.write(cab)
        escreve_termicas()
        # Escreve a linha de terminação
        arq.write(f" {BlocoUTEsAdTerm.str_fim}\n")


class LeituraAdTerm(Leitura):
    """
    Realiza a leitura do arquivo `adterm.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `adterm.dat`, construindo
    um objeto `AdTerm` cujas informações são as mesmas do adterm.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self,
                 diretorio: str):
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Cria a lista de blocos a serem lidos no arquivo adterm.dat.
        """
        return [BlocoUTEsAdTerm()]



# Rotinas de testes associadas ao arquivo adterm.dat do NEWAVE
from inewave.newave.cadic import CAdic


cad = CAdic.le_arquivo("tests/_arquivos")


def test_leitura():
    assert len(cad.cargas_adicionais) > 0


def test_escrita_e_leitura():
    cad.escreve_arquivo("tests/_saidas")
    cad2 = CAdic.le_arquivo("tests/_saidas")
    assert cad == cad2


def test_eq_cadic():
    cad2 = CAdic.le_arquivo("tests/_arquivos")
    assert cad == cad2


def test_neq_cadic():
    cad2 = CAdic.le_arquivo("tests/_arquivos")
    cad2.cargas_adicionais.iloc[0, 0] = -1
    assert cad2 != cad
    assert cad2 is not None



from inewave._utils.arquivo import Arquivo
from inewave._utils.dadosarquivo import DadosArquivo
from inewave._utils.escrita import Escrita
from inewave.newave.modelos.caso import BlocoCaso, LeituraCaso


class Caso(Arquivo):
    """
    Armazena os dados de entrada do NEWAVE referentes ao arquivo
    `caso.dat`.

    Esta classe lida com informações de entrada do NEWAVE e
    que deve se referir ao caminho do `arquivos.dat`.

    **Parâmetros**

    - caso: `str`

    """

    def __init__(self,
                 dados: DadosArquivo):
        super().__init__(dados)
        val = True
        msg = "Erro na criação de Caso: "
        if len(dados.blocos) == 1:
            bloco = dados.blocos[0]
            if isinstance(bloco, BlocoCaso):
                self.__bloco = bloco
            else:
                msg += (f"O bloco deve ser do tipo {BlocoCaso}, " +
                        f"mas foi fornecido do tipo {type(bloco)}")
                val = False
        else:
            msg += "Deve ser fornecido exatamente 1 bloco para Caso"
            val = False
        if not val:
            raise TypeError(msg)

    @classmethod
    def le_arquivo(cls,
                   diretorio: str,
                   nome_arquivo="caso.dat") -> 'Caso':
        """
        """
        leitor = LeituraCaso(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    def escreve_arquivo(self,
                        diretorio: str,
                        nome_arquivo="caso.dat"):
        """
        """
        escritor = Escrita(diretorio)
        escritor.escreve_arquivo(self._dados,
                                 nome_arquivo)

    @property
    def arquivo(self) -> str:
        """
        Caminho para o arquivo `arquivos.dat` de entrada do NEWAVE.

        **Retorna**

        `str`

        **Sobre**

        Retorna o caminho completo (unix-like) para o arquivo que direciona
        para os demais dados de entrada do NEWAVE.
        """
        return self.__bloco.dados



from abc import abstractmethod
from .dadosarquivo import DadosArquivo


class Arquivo:
    """
    """
    def __init__(self,
                 dados: DadosArquivo) -> None:
        self._dados = dados

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre arquivos avalia os dados interpretados e
        também os comentários externos.
        """
        if not isinstance(o, Arquivo):
            return False
        arquivos: Arquivo = o
        dif = False
        for (i1, l1), (i2, l2) in zip(self._linhas_fora_blocos.items(),
                                      arquivos._linhas_fora_blocos.items()):
            if i1 != i2 or l1 != l2:
                dif = True
                break
        for b1, b2 in zip(self._blocos,
                          arquivos._blocos):
            if b1 != b2:
                dif = True
                break
        return not dif

    @property
    def _linhas_fora_blocos(self):
        return self._dados.linhas_fora_blocos

    @property
    def _blocos(self):
        return self._dados.blocos

    @classmethod
    @abstractmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="") -> 'Arquivo':
        pass

    @abstractmethod
    def escreve_arquivo(self, diretorio: str, nome_arquivo=""):
        pass



from inewave.config import SUBMERCADOS, MESES, MESES_DF, MAX_ANOS_ESTUDO
from inewave._utils.bloco import Bloco
from inewave._utils.leitura import Leitura
from inewave._utils.registros import RegistroAn, RegistroIn, RegistroFn

from typing import List, IO
import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class BlocoCargasAdicionaisSubsistema(Bloco):
    """
    Bloco com informações de cargas adicionais por mês/ano
    e por subsistema.
    """
    str_inicio = ""
    str_fim = "999"

    def __init__(self):

        super().__init__(BlocoCargasAdicionaisSubsistema.str_inicio,
                         "",
                         True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoCargasAdicionaisSubsistema):
            return False
        bloco: BlocoCargasAdicionaisSubsistema = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):

        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            df.columns = MESES_DF
            df["Ano"] = anos
            df["Num. Subsistema"] = subsistema
            df["Nome Subsistema"] = nome_subsistema
            df["Razão Carga"] = razao_carga
            df = df[["Ano", "Num. Subsistema",
                     "Nome Subsistema", "Razão Carga"] + MESES_DF]
            return df

        # Variáveis auxiliares
        reg_subsis = RegistroIn(3)
        reg_nome_subsis = RegistroAn(12)
        reg_razao_carga = RegistroAn(12)
        reg_ano = RegistroAn(4)
        reg_carga = RegistroFn(7)
        # Pula uma linha, com cabeçalhos
        arq.readline()
        i = 0
        anos = []
        subsistema = []
        nome_subsistema = []
        subsistema_atual = 0
        nome_subsistema_atual = ""
        razao_carga = []
        razao_carga_atual = ""
        tabela = np.zeros((MAX_ANOS_ESTUDO * len(SUBMERCADOS),
                          len(MESES)))
        while True:
            # Verifica se o arquivo acabou
            linha: str = arq.readline()
            if BlocoCargasAdicionaisSubsistema.str_fim == linha.strip():
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            # Senão, lê mais uma linha
            if len(linha.strip()) < 40:
                subsistema_atual = reg_subsis.le_registro(linha, 1)
                nome_subsistema_atual = reg_nome_subsis.le_registro(linha, 6)
                razao_carga_atual = reg_razao_carga.le_registro(linha, 20)
            else:
                # Ano
                anos.append(reg_ano.le_registro(linha, 0))
                # Subsistema
                subsistema.append(subsistema_atual)
                nome_subsistema.append(nome_subsistema_atual)
                razao_carga.append(razao_carga_atual)
                # Limites
                ci = 7
                nc = 7
                for j in range(len(MESES)):
                    cf = ci + nc
                    if len(linha[ci:cf].strip()) > 0:
                        tabela[i, j] = reg_carga.le_registro(linha, ci)
                    ci = cf + 1
                i += 1

    # Override
    def escreve(self, arq: IO):

        def escreve_cargas():
            lin_tab = self._dados.shape[0]
            subsistema_anterior = 0
            nome_anterior = ""
            razao_anterior = ""
            for i in range(lin_tab):
                linha = ""
                # Subsistema
                subsistema = self._dados.iloc[i, 1]
                nome = self._dados.iloc[i, 2]
                razao = self._dados.iloc[i, 3]
                if any([
                        subsistema != subsistema_anterior,
                        nome != nome_anterior,
                        razao != razao_anterior
                       ]):
                    subsistema_anterior = subsistema
                    nome_anterior = nome
                    razao_anterior = razao
                    linha = (" " + str(subsistema).rjust(3) +
                             "  " + str(nome).ljust(12) +
                             "  " + str(razao).rjust(12))
                    arq.write(linha + "\n")
                    subsistema_anterior = subsistema
                # Mercados de cada mês
                linha = f"{self._dados.iloc[i, 0].ljust(4)}  "
                for j in range(len(MESES)):
                    v = self._dados.iloc[i, j + 4]
                    if v == 0:
                        linha += "        "
                    else:
                        linha += " " + "{:6.0f}.".format(v).rjust(7)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        arq.write(" XXX\n")
        cab = ("       XXXJAN. XXXFEV. XXXMAR. XXXABR. XXXMAI. XXXJUN." +
               " XXXJUL. XXXAGO. XXXSET. XXXOUT. XXXNOV. XXXDEZ.\n")
        arq.write(cab)
        escreve_cargas()
        # Escreve a linha de terminação
        arq.write(f" {BlocoCargasAdicionaisSubsistema.str_fim}\n")


class LeituraCAdic(Leitura):
    """
    Realiza a leitura do arquivo `cadic.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `cadic.dat`, construindo
    um objeto `CAdic` cujas informações são as mesmas do cadic.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self,
                 diretorio: str):
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Cria a lista de blocos a serem lidos no arquivo cadic.dat.
        """
        return [BlocoCargasAdicionaisSubsistema()]
